#!/usr/bin/python3

import datetime, os, unittest

from slippi import Game
from slippi.id import InGameCharacter, CSSCharacter, Stage
from slippi.event import Start, End, Frame, Buttons, Triggers, Position


BPhys = Buttons.Physical
BLog = Buttons.Logical


def norm(f):
    return 1 if f > 0.01 else -1 if f < -0.01 else 0


class TestGame(unittest.TestCase):
    def _game(self, name):
        return Game(os.path.join(os.path.dirname(__file__), 'replays', name+'.slp'))

    def _stick_seq(self, game):
        pass

    # not used yet because the recorder currently puts bogus values in triggers.physical
    def _trigger_seq(self, game):
        last_triggers = None
        trigger_seq = []
        for frame in game.frames:
            t = frame.ports[0].leader.pre.triggers
            t = Triggers(norm(t.logical), norm(t.physical.l), norm(t.physical.r))
            if (t.logical or t.physical.l or t.physical.r) and t != last_triggers:
                trigger_seq.append(t)
            last_triggers = t
        return trigger_seq

    def _button_seq(self, game):
        last_buttons = None
        button_seq = []
        for frame in game.frames:
            b = frame.ports[0].leader.pre.buttons
            if (b.logical or b.physical) and b != last_buttons:
                button_seq.append(b)
            last_buttons = b
        return button_seq

    def test_slippi_old_version(self):
        game = self._game('v0.1')
        self.assertEqual(game.start.slippi.version, Start.Slippi.Version(0,1,0,0))
        self.assertEqual(game.metadata.duration, None)
        self.assertEqual(game.metadata.players, (None, None, None, None))
        self.assertEqual(game.start.players[0].character, CSSCharacter.FOX)
        self.assertEqual(game.start.players[1].character, CSSCharacter.GANONDORF)

    def test_game(self):
        game = self._game('game')

        self.assertEqual(game.metadata, Game.Metadata._parse({
            'startAt': '2018-06-22T07:52:59Z',
            'lastFrame': 5085,
            'playedOn': 'dolphin',
            'players': {
                '0': {'characters': {InGameCharacter.MARTH: 5209}},
                '1': {'characters': {InGameCharacter.FOX: 5209}}}}))
        self.assertEqual(game.metadata, Game.Metadata(
            date=datetime.datetime(2018, 6, 22, 7, 52, 59, 0, datetime.timezone.utc),
            duration=5086,
            platform=Game.Metadata.Platform.DOLPHIN,
            players=(
                Game.Metadata.Player({InGameCharacter.MARTH: 5086}),
                Game.Metadata.Player({InGameCharacter.FOX: 5086}),
                None, None)))

        self.assertEqual(game.start, Start(
            is_teams=False,
            random_seed=3803194226,
            slippi=Start.Slippi(Start.Slippi.Version(1,0,0,0)),
            stage=Stage.YOSHIS_STORY,
            players=(
                Start.Player(character=CSSCharacter.MARTH, type=Start.Player.Type.HUMAN, stocks=4, costume=3, team=None, ucf=Start.Player.UCF(False, False)),
                Start.Player(character=CSSCharacter.FOX, type=Start.Player.Type.CPU, stocks=4, costume=0, team=None, ucf=Start.Player.UCF(False, False)),
                None, None)))

        self.assertEqual(game.end, End(End.Method.CONCLUSIVE))

        self.assertEqual(game.metadata.duration, len(game.frames))

    def test_ics(self):
        game = self._game('ics')
        self.assertEqual(game.metadata.players[0].characters, {
            InGameCharacter.NANA: 221,
            InGameCharacter.POPO: 221})
        self.assertEqual(game.start.players[0].character, CSSCharacter.ICE_CLIMBERS)
        self.assertIsNotNone(game.frames[0].ports[0].follower)

    def test_ucf(self):
        self.assertEqual(self._game('shield_drop').start.players[0].ucf, Start.Player.UCF(dash_back=False, shield_drop=True))
        self.assertEqual(self._game('dash_back').start.players[0].ucf, Start.Player.UCF(dash_back=True, shield_drop=False))

    def test_buttons_lrzs(self):
        game = self._game('buttons-lrzs')
        self.assertEqual(self._button_seq(game), [
            Buttons(BLog.TRIGGER_ANALOG, BPhys.NONE),
            Buttons(BLog.TRIGGER_ANALOG|BLog.L, BPhys.L),
            Buttons(BLog.TRIGGER_ANALOG, BPhys.NONE),
            Buttons(BLog.TRIGGER_ANALOG|BLog.R, BPhys.R),
            Buttons(BLog.TRIGGER_ANALOG|BLog.A|BLog.Z, BPhys.Z),
            Buttons(BLog.START, BPhys.START)])

    def test_buttons_abxy(self):
        game = self._game('buttons-abxy')
        self.assertEqual(self._button_seq(game), [
            Buttons(BLog.A, BPhys.A),
            Buttons(BLog.B, BPhys.B),
            Buttons(BLog.X, BPhys.X),
            Buttons(BLog.Y, BPhys.Y)])

    def test_dpad_udlr(self):
        game = self._game('dpad-udlr')
        self.assertEqual(self._button_seq(game), [
            Buttons(BLog.DPAD_UP, BPhys.DPAD_UP),
            Buttons(BLog.DPAD_DOWN, BPhys.DPAD_DOWN),
            Buttons(BLog.DPAD_LEFT, BPhys.DPAD_LEFT),
            Buttons(BLog.DPAD_RIGHT, BPhys.DPAD_RIGHT)])

    def test_cstick_udlr(self):
        game = self._game('cstick-udlr')
        self.assertEqual(self._button_seq(game), [
            Buttons(BLog.CSTICK_UP, BPhys.NONE),
            Buttons(BLog.CSTICK_DOWN, BPhys.NONE),
            Buttons(BLog.CSTICK_LEFT, BPhys.NONE),
            Buttons(BLog.CSTICK_RIGHT, BPhys.NONE)])

    def test_joystick_udlr(self):
        game = self._game('joystick-udlr')
        self.assertEqual(self._button_seq(game), [
            Buttons(BLog.JOYSTICK_UP, BPhys.NONE),
            Buttons(BLog.JOYSTICK_DOWN, BPhys.NONE),
            Buttons(BLog.JOYSTICK_LEFT, BPhys.NONE),
            Buttons(BLog.JOYSTICK_RIGHT, BPhys.NONE)])


if __name__ == '__main__':
    unittest.main()



import inspect, struct


def _real_attrs(obj):
    attrs = ((attr, obj.__getattribute__(attr)) for attr in dir(obj)
             if not attr.startswith('_'))
    return (a for a in attrs if not inspect.isclass(a[1]))


def dict_repr(obj):
    return '<%s: %s>' % (obj.__class__.__qualname__,
                         ', '.join('%s: %s' % a for a in _real_attrs(obj)))


def unpack(fmt, stream):
    fmt = '>' + fmt
    size = struct.calcsize(fmt)
    bytes = stream.read(size)
    if not bytes:
        raise EofException()
    return struct.unpack(fmt, bytes)


class EofException(Exception):
    pass



from enum import IntEnum, IntFlag

from slippi.util import *
from slippi.id import *


class EventType(IntEnum):
    EVENT_PAYLOADS = 0x35
    GAME_START = 0x36
    PRE_FRAME = 0x37
    POST_FRAME = 0x38
    GAME_END = 0x39


class Start:
    """Information used to initialize the game such as the game mode, settings, characters & stage."""

    def __init__(self, is_teams, players, random_seed, slippi, stage):
        self.is_teams = is_teams #: :py:class:`bool`: True if this was a teams game
        self.players = players #: tuple(optional(:py:class:`Player`)): Players in this game by port (port 1 is at index 0; empty ports will contain None)
        self.random_seed = random_seed #: :py:class:`int`: Random seed before the game start
        self.slippi = slippi #: :py:class:`Slippi`: Information about the Slippi recorder that generated this replay
        self.stage = stage #: :py:class:`slippi.id.Stage`: Stage on which this game was played

    @classmethod
    def _parse(cls, stream):
        slippi = cls.Slippi._parse(stream)

        stream.read(8)
        (is_teams,) = unpack('?', stream)

        stream.read(5)
        (stage,) = unpack('H', stream)
        stage = Stage(stage)

        stream.read(80)
        players = []
        for i in range(4):
            (character, type, stocks, costume) = unpack('BBBB', stream)
            character = CSSCharacter(character)

            stream.read(5)
            (team,) = unpack('B', stream)
            team = cls.Player.Team(team) if is_teams else None

            stream.read(26)
            try:
                player = cls.Player(character=character, type=cls.Player.Type(type), stocks=stocks, costume=costume, team=team)
            except ValueError: player = None

            players.append(player)

        stream.read(72)
        (random_seed,) = unpack('L', stream)

        try:
            # added: 1.0.0.0
            for i in range(4):
                (dash_back, shield_drop) = unpack('LL', stream)
                if players[i]:
                    players[i].ucf = cls.Player.UCF(dash_back != 0, shield_drop != 0)
        except EofException: pass

        return cls(is_teams=is_teams, players=tuple(players), random_seed=random_seed, slippi=slippi, stage=stage)

    __repr__ = dict_repr

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.is_teams == other.is_teams and self.players == other.players and self.random_seed == other.random_seed and self.slippi == other.slippi and self.stage is other.stage


    class Slippi:
        """Information about the Slippi recorder that generated this replay."""

        def __init__(self, version):
            self.version = version #: :py:class:`Version`: Slippi version number

        @classmethod
        def _parse(cls, stream):
            return cls(cls.Version(*unpack('BBBB', stream)))

        __repr__ = dict_repr

        def __eq__(self, other):
            if not isinstance(other, self.__class__):
                return NotImplemented
            return self.version == other.version


        class Version:
            def __init__(self, major, minor, revision, build):
                self.major = major #: :py:class:`int`:
                self.minor = minor #: :py:class:`int`:
                self.revision = revision #: :py:class:`int`:
                self.build = build #: :py:class:`int`:

            def __repr__(self):
                return '%d.%d.%d.%d' % (self.major, self.minor, self.revision, self.build)

            def __eq__(self, other):
                if not isinstance(other, self.__class__):
                    return NotImplemented
                return self.major == other.major and self.minor == other.minor and self.revision == other.revision and self.build == other.build


    class Player:
        def __init__(self, character, type, stocks, costume, team, ucf = None):
            self.character = character #: :py:class:`slippi.id.CSSCharacter`: Character selected
            self.type = type #: :py:class:`Type`: Player type (human/cpu)
            self.stocks = stocks #: :py:class:`int`: Starting stock count
            self.costume = costume #: :py:class:`int`: Costume ID
            self.team = team #: optional(:py:class:`Team`): Team, if this was a teams game
            self.ucf = ucf or self.UCF() #: :py:class:`UCF`: UCF feature toggles

        __repr__ = dict_repr

        def __eq__(self, other):
            if not isinstance(other, self.__class__):
                return NotImplemented
            return self.character is other.character and self.type is other.type and self.stocks == other.stocks and self.costume == other.costume and self.team is other.team and self.ucf == other.ucf


        class Type(IntEnum):
            HUMAN = 0
            CPU = 1


        class Team(IntEnum):
            RED = 0
            BLUE = 1
            GREEN = 2


        class UCF:
            def __init__(self, dash_back = False, shield_drop = False):
                self.dash_back = bool(dash_back) #: :py:class:`bool`: True if UCF dashback was enabled
                self.shield_drop = bool(shield_drop) #: :py:class:`bool`: True if UCF shield drop was enabled

            __repr__ = dict_repr

            def __eq__(self, other):
                if not isinstance(other, self.__class__):
                    return NotImplemented
                return self.dash_back == other.dash_back and self.shield_drop == other.shield_drop


class End:
    """Information about the end of the game."""

    def __init__(self, method):
        self.method = method #: :py:class:`Method`: How the game ended

    @classmethod
    def _parse(cls, stream):
        (method,) = unpack('B', stream)
        return cls(cls.Method(method))

    __repr__ = dict_repr

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.method is other.method


    class Method(IntEnum):
        INCONCLUSIVE = 0
        CONCLUSIVE = 3


class Frame:
    """A single frame of the game. Includes data for all characters."""

    def __init__(self):
        self.ports = [None, None, None, None]
        """tuple(optional(:py:class:`Port`)): Frame data for each port (port 1 is at index 0; empty ports will contain None)."""

    def _finalize(self):
        self.ports = tuple(self.ports)

    __repr__ = dict_repr


    class Port:
        """Frame data for a given port. Can include two characters' frame data (ICs)."""

        def __init__(self):
            self.leader = self.Data() #: :py:class:`Data`: Frame data for the controlled character
            self.follower = None #: optional(:py:class:`Data`): Frame data for the follower (Nana), if any

        __repr__ = dict_repr


        class Data:
            """Frame data for a given character. Includes both pre-frame and post-frame data."""

            def __init__(self):
                self.pre = None #: :py:class:`Pre`: Pre-frame update data
                self.post = None #: :py:class:`Post`: Post-frame update data

            __repr__ = dict_repr


            class Pre:
                """Pre-frame update data, required to reconstruct a replay. Information is collected right before controller inputs are used to figure out the character's next action."""

                def __init__(self, stream):
                    (random_seed, state, position_x, position_y, direction, joystick_x, joystick_y, cstick_x, cstick_y, trigger_logical, buttons_logical, buttons_physical, trigger_physical_l, trigger_physical_r) = unpack('LHffffffffLHff', stream)
                    self.state = ActionState(state) #: :py:class:`slippi.id.ActionState`: Character's action state (useful for stats)
                    self.position = Position(position_x, position_y) #: :py:class:`Position`: Character's position
                    self.direction = Direction(direction) #: :py:class:`Direction`: Direction the character is facing
                    self.joystick = Position(joystick_x, joystick_y) #: :py:class:`Position`: Processed analog joystick position
                    self.cstick = Position(cstick_x, cstick_y) #: :py:class:`Position`: Processed analog c-stick position
                    self.triggers = Triggers(trigger_logical, trigger_physical_l, trigger_physical_r) #: :py:class:`Triggers`: Trigger state
                    self.buttons = Buttons(buttons_logical, buttons_physical) #: :py:class:`Buttons`: Button state
                    self.random_seed = random_seed #: :py:class:`int`: Random seed at this point

                __repr__ = dict_repr


            class Post:
                """Post-frame update data, for making decisions about game states (such as computing stats). Information is collected at the end of collision detection, which is the last consideration of the game engine."""

                def __init__(self, stream):
                    (character, state, position_x, position_y, direction, damage, shield, last_attack_landed, combo_count, last_hit_by, stocks) = unpack('BHfffffBBBB', stream)
                    state_age = None
                    try:
                        # added: 0.2.0.0
                        (state_age,) = unpack('f', stream)
                    except EofException: pass
                    self.character = InGameCharacter(character) #: :py:class:`slippi.id.InGameCharacter`: In-game character (can only change for Zelda/Sheik). Check on first frame to determine if Zelda started as Sheik
                    self.state = ActionState(state) #: :py:class:`slippi.id.ActionState`: Character's action state (useful for stats)
                    self.state_age = state_age #: :py:class:`float`: Number of frames action state has been active. Can have a fractional component for certain actions
                    self.position = Position(position_x, position_y) #: :py:class:`Position`: Character's position
                    self.direction = Direction(direction) #: :py:class:`Direction`: Direction the character is facing
                    self.damage = damage #: :py:class:`float`: Current damage percent
                    self.shield = shield #: :py:class:`float`: Current size of shield
                    self.stocks = stocks #: :py:class:`int`: Number of stocks remaining
                    self.last_attack_landed = Attack(last_attack_landed) if last_attack_landed else None #: optional(:py:class:`Attack`): Last attack that this character landed
                    self.last_hit_by = last_hit_by if last_hit_by < 4 else None #: optional(:py:class:`int`): Port of character that last hit this character
                    self.combo_count = combo_count #: :py:class:`int`: Combo count as defined by the game

                __repr__ = dict_repr


    # This class is only used temporarily while parsing frame data.
    class Event:
        def __init__(self, id, data):
            self.id = id
            self.data = data


        class Id:
            def __init__(self, stream):
                (self.frame, self.port, self.is_follower) = unpack('iB?', stream)
                pass


class Position:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.x == other.x and self.y == other.y


class Direction(IntEnum):
    LEFT = -1
    RIGHT = 1


class Attack(IntEnum):
    OTHER = 1
    JAB = 2
    JAB2 = 3
    JAB3 = 4
    DASH_ATTACK = 6
    FTILT = 7
    UTILT = 8
    DTILT = 9
    FSMASH = 10
    USMASH = 11
    DSMASH = 12
    NAIR = 13
    FAIR = 14
    BAIR = 15
    UAIR = 16
    DAIR = 17
    NEUTRAL_B = 18
    SIDE_B = 19
    UP_B = 20
    DOWN_B = 21
    GETUP_ATTACK = 50
    GETUP_ATTACK_SLOW = 51
    GRAB_RELEASE = 52
    FTHROW = 53
    BTHROW = 54
    UTHROW = 55
    DTHROW = 56
    LEDGE_ATTACK_SLOW = 61
    LEDGE_ATTACK = 62


class Triggers:
    def __init__(self, logical, physical_x, physical_y):
        self.logical = logical #: :py:class:`float`: Processed analog trigger position
        self.physical = self.Physical(physical_x, physical_y) #: :py:class:`Physical`: physical analog trigger positions (useful for APM)

    __repr__ = dict_repr

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return other.logical == self.logical and other.physical == self.physical


    class Physical:
        def __init__(self, l, r):
            self.l = l
            self.r = r

        __repr__ = dict_repr

        def __eq__(self, other):
            if not isinstance(other, self.__class__):
                return NotImplemented
            # Should we add an epsilon to these comparisons? When are people going to be comparing trigger states for equality, other than in our tests?
            return other.l == self.l and other.r == self.r


class Buttons:
    def __init__(self, logical, physical):
        self.logical = self.Logical(logical) #: :py:class:`Logical`: Processed button-state bitmask
        self.physical = self.Physical(physical) #: :py:class:`Physical`: Physical button-state bitmask

    def __eq__(self, other):
        if not isinstance(other, Buttons):
            return NotImplemented
        return other.logical is self.logical and other.physical is self.physical

    __repr__ = dict_repr


    class Logical(IntFlag):
        TRIGGER_ANALOG = 2**31
        CSTICK_RIGHT = 2**23
        CSTICK_LEFT = 2**22
        CSTICK_DOWN = 2**21
        CSTICK_UP = 2**20
        JOYSTICK_RIGHT = 2**19
        JOYSTICK_LEFT = 2**18
        JOYSTICK_DOWN = 2**17
        JOYSTICK_UP = 2**16
        START = 2**12
        Y = 2**11
        X = 2**10
        B = 2**9
        A = 2**8
        L = 2**6
        R = 2**5
        Z = 2**4
        DPAD_UP = 2**3
        DPAD_DOWN = 2**2
        DPAD_RIGHT = 2**1
        DPAD_LEFT = 2**0
        NONE = 0


    class Physical(IntFlag):
        START = 2**12
        Y = 2**11
        X = 2**10
        B = 2**9
        A = 2**8
        L = 2**6
        R = 2**5
        Z = 2**4
        DPAD_UP = 2**3
        DPAD_DOWN = 2**2
        DPAD_RIGHT = 2**1
        DPAD_LEFT = 2**0
        NONE = 0

        def pressed(self):
            """Returns a list of all buttons being pressed."""
            pressed = []
            for b in self.__class__:
                if self & b:
                    pressed.append(b)
            return pressed



# These IDs (and other very useful info for this project) came from the SSBM Data Sheet: https://docs.google.com/spreadsheets/d/1JX2w-r2fuvWuNgGb6D3Cs4wHQKLFegZe2jhbBuIhCG8

from enum import IntEnum


class ActionState(IntEnum):
    DEAD_DOWN = 0
    DEAD_LEFT = 1
    DEAD_RIGHT = 2
    DEAD_UP = 3
    DEAD_UP_STAR = 4
    DEAD_UP_STAR_ICE = 5
    DEAD_UP_FALL = 6
    DEAD_UP_FALL_HIT_CAMERA = 7
    DEAD_UP_FALL_HIT_CAMERA_FLAT = 8
    DEAD_UP_FALL_ICE = 9
    DEAD_UP_FALL_HIT_CAMERA_ICE = 10
    SLEEP = 11
    REBIRTH = 12
    REBIRTH_WAIT = 13
    WAIT = 14
    WALK_SLOW = 15
    WALK_MIDDLE = 16
    WALK_FAST = 17
    TURN = 18
    TURN_RUN = 19
    DASH = 20
    RUN = 21
    RUN_DIRECT = 22
    RUN_BRAKE = 23
    KNEE_BEND = 24
    JUMP_F = 25
    JUMP_B = 26
    JUMP_AERIAL_F = 27
    JUMP_AERIAL_B = 28
    FALL = 29
    FALL_F = 30
    FALL_B = 31
    FALL_AERIAL = 32
    FALL_AERIAL_F = 33
    FALL_AERIAL_B = 34
    FALL_SPECIAL = 35
    FALL_SPECIAL_F = 36
    FALL_SPECIAL_B = 37
    DAMAGE_FALL = 38
    SQUAT = 39
    SQUAT_WAIT = 40
    SQUAT_RV = 41
    LANDING = 42
    LANDING_FALL_SPECIAL = 43
    ATTACK11 = 44
    ATTACK12 = 45
    ATTACK13 = 46
    ATTACK100START = 47
    ATTACK100LOOP = 48
    ATTACK100END = 49
    ATTACK_DASH = 50
    ATTACK_S3HI = 51
    ATTACK_S3HI_S = 52
    ATTACK_S3S = 53
    ATTACK_S3LW_S = 54
    ATTACK_S3LW = 55
    ATTACK_HI3 = 56
    ATTACK_LW3 = 57
    ATTACK_S4HI = 58
    ATTACK_S4HI_S = 59
    ATTACK_S4S = 60
    ATTACK_S4LW_S = 61
    ATTACK_S4LW = 62
    ATTACK_HI4 = 63
    ATTACK_LW4 = 64
    ATTACK_AIR_N = 65
    ATTACK_AIR_F = 66
    ATTACK_AIR_B = 67
    ATTACK_AIR_HI = 68
    ATTACK_AIR_LW = 69
    LANDING_AIR_N = 70
    LANDING_AIR_F = 71
    LANDING_AIR_B = 72
    LANDING_AIR_HI = 73
    LANDING_AIR_LW = 74
    DAMAGE_HI1 = 75
    DAMAGE_HI2 = 76
    DAMAGE_HI3 = 77
    DAMAGE_N1 = 78
    DAMAGE_N2 = 79
    DAMAGE_N3 = 80
    DAMAGE_LW1 = 81
    DAMAGE_LW2 = 82
    DAMAGE_LW3 = 83
    DAMAGE_AIR1 = 84
    DAMAGE_AIR2 = 85
    DAMAGE_AIR3 = 86
    DAMAGE_FLY_HI = 87
    DAMAGE_FLY_N = 88
    DAMAGE_FLY_LW = 89
    DAMAGE_FLY_TOP = 90
    DAMAGE_FLY_ROLL = 91
    LIGHT_GET = 92
    HEAVY_GET = 93
    LIGHT_THROW_F = 94
    LIGHT_THROW_B = 95
    LIGHT_THROW_HI = 96
    LIGHT_THROW_LW = 97
    LIGHT_THROW_DASH = 98
    LIGHT_THROW_DROP = 99
    LIGHT_THROW_AIR_F = 100
    LIGHT_THROW_AIR_B = 101
    LIGHT_THROW_AIR_HI = 102
    LIGHT_THROW_AIR_LW = 103
    HEAVY_THROW_F = 104
    HEAVY_THROW_B = 105
    HEAVY_THROW_HI = 106
    HEAVY_THROW_LW = 107
    LIGHT_THROW_F4 = 108
    LIGHT_THROW_B4 = 109
    LIGHT_THROW_HI4 = 110
    LIGHT_THROW_LW4 = 111
    LIGHT_THROW_AIR_F4 = 112
    LIGHT_THROW_AIR_B4 = 113
    LIGHT_THROW_AIR_HI4 = 114
    LIGHT_THROW_AIR_LW4 = 115
    HEAVY_THROW_F4 = 116
    HEAVY_THROW_B4 = 117
    HEAVY_THROW_HI4 = 118
    HEAVY_THROW_LW4 = 119
    SWORD_SWING1 = 120
    SWORD_SWING3 = 121
    SWORD_SWING4 = 122
    SWORD_SWING_DASH = 123
    BAT_SWING1 = 124
    BAT_SWING3 = 125
    BAT_SWING4 = 126
    BAT_SWING_DASH = 127
    PARASOL_SWING1 = 128
    PARASOL_SWING3 = 129
    PARASOL_SWING4 = 130
    PARASOL_SWING_DASH = 131
    HARISEN_SWING1 = 132
    HARISEN_SWING3 = 133
    HARISEN_SWING4 = 134
    HARISEN_SWING_DASH = 135
    STAR_ROD_SWING1 = 136
    STAR_ROD_SWING3 = 137
    STAR_ROD_SWING4 = 138
    STAR_ROD_SWING_DASH = 139
    LIP_STICK_SWING1 = 140
    LIP_STICK_SWING3 = 141
    LIP_STICK_SWING4 = 142
    LIP_STICK_SWING_DASH = 143
    ITEM_PARASOL_OPEN = 144
    ITEM_PARASOL_FALL = 145
    ITEM_PARASOL_FALL_SPECIAL = 146
    ITEM_PARASOL_DAMAGE_FALL = 147
    L_GUN_SHOOT = 148
    L_GUN_SHOOT_AIR = 149
    L_GUN_SHOOT_EMPTY = 150
    L_GUN_SHOOT_AIR_EMPTY = 151
    FIRE_FLOWER_SHOOT = 152
    FIRE_FLOWER_SHOOT_AIR = 153
    ITEM_SCREW = 154
    ITEM_SCREW_AIR = 155
    DAMAGE_SCREW = 156
    DAMAGE_SCREW_AIR = 157
    ITEM_SCOPE_START = 158
    ITEM_SCOPE_RAPID = 159
    ITEM_SCOPE_FIRE = 160
    ITEM_SCOPE_END = 161
    ITEM_SCOPE_AIR_START = 162
    ITEM_SCOPE_AIR_RAPID = 163
    ITEM_SCOPE_AIR_FIRE = 164
    ITEM_SCOPE_AIR_END = 165
    ITEM_SCOPE_START_EMPTY = 166
    ITEM_SCOPE_RAPID_EMPTY = 167
    ITEM_SCOPE_FIRE_EMPTY = 168
    ITEM_SCOPE_END_EMPTY = 169
    ITEM_SCOPE_AIR_START_EMPTY = 170
    ITEM_SCOPE_AIR_RAPID_EMPTY = 171
    ITEM_SCOPE_AIR_FIRE_EMPTY = 172
    ITEM_SCOPE_AIR_END_EMPTY = 173
    LIFT_WAIT = 174
    LIFT_WALK1 = 175
    LIFT_WALK2 = 176
    LIFT_TURN = 177
    GUARD_ON = 178
    GUARD = 179
    GUARD_OFF = 180
    GUARD_SET_OFF = 181
    GUARD_REFLECT = 182
    DOWN_BOUND_U = 183
    DOWN_WAIT_U = 184
    DOWN_DAMAGE_U = 185
    DOWN_STAND_U = 186
    DOWN_ATTACK_U = 187
    DOWN_FOWARD_U = 188
    DOWN_BACK_U = 189
    DOWN_SPOT_U = 190
    DOWN_BOUND_D = 191
    DOWN_WAIT_D = 192
    DOWN_DAMAGE_D = 193
    DOWN_STAND_D = 194
    DOWN_ATTACK_D = 195
    DOWN_FOWARD_D = 196
    DOWN_BACK_D = 197
    DOWN_SPOT_D = 198
    PASSIVE = 199
    PASSIVE_STAND_F = 200
    PASSIVE_STAND_B = 201
    PASSIVE_WALL = 202
    PASSIVE_WALL_JUMP = 203
    PASSIVE_CEIL = 204
    SHIELD_BREAK_FLY = 205
    SHIELD_BREAK_FALL = 206
    SHIELD_BREAK_DOWN_U = 207
    SHIELD_BREAK_DOWN_D = 208
    SHIELD_BREAK_STAND_U = 209
    SHIELD_BREAK_STAND_D = 210
    FURA_FURA = 211
    CATCH = 212
    CATCH_PULL = 213
    CATCH_DASH = 214
    CATCH_DASH_PULL = 215
    CATCH_WAIT = 216
    CATCH_ATTACK = 217
    CATCH_CUT = 218
    THROW_F = 219
    THROW_B = 220
    THROW_HI = 221
    THROW_LW = 222
    CAPTURE_PULLED_HI = 223
    CAPTURE_WAIT_HI = 224
    CAPTURE_DAMAGE_HI = 225
    CAPTURE_PULLED_LW = 226
    CAPTURE_WAIT_LW = 227
    CAPTURE_DAMAGE_LW = 228
    CAPTURE_CUT = 229
    CAPTURE_JUMP = 230
    CAPTURE_NECK = 231
    CAPTURE_FOOT = 232
    ESCAPE_F = 233
    ESCAPE_B = 234
    ESCAPE = 235
    ESCAPE_AIR = 236
    REBOUND_STOP = 237
    REBOUND = 238
    THROWN_F = 239
    THROWN_B = 240
    THROWN_HI = 241
    THROWN_LW = 242
    THROWN_LW_WOMEN = 243
    PASS = 244
    OTTOTTO = 245
    OTTOTTO_WAIT = 246
    FLY_REFLECT_WALL = 247
    FLY_REFLECT_CEIL = 248
    STOP_WALL = 249
    STOP_CEIL = 250
    MISS_FOOT = 251
    CLIFF_CATCH = 252
    CLIFF_WAIT = 253
    CLIFF_CLIMB_SLOW = 254
    CLIFF_CLIMB_QUICK = 255
    CLIFF_ATTACK_SLOW = 256
    CLIFF_ATTACK_QUICK = 257
    CLIFF_ESCAPE_SLOW = 258
    CLIFF_ESCAPE_QUICK = 259
    CLIFF_JUMP_SLOW1 = 260
    CLIFF_JUMP_SLOW2 = 261
    CLIFF_JUMP_QUICK1 = 262
    CLIFF_JUMP_QUICK2 = 263
    APPEAL_R = 264
    APPEAL_L = 265
    SHOULDERED_WAIT = 266
    SHOULDERED_WALK_SLOW = 267
    SHOULDERED_WALK_MIDDLE = 268
    SHOULDERED_WALK_FAST = 269
    SHOULDERED_TURN = 270
    THROWN_F_F = 271
    THROWN_F_B = 272
    THROWN_F_HI = 273
    THROWN_F_LW = 274
    CAPTURE_CAPTAIN = 275
    CAPTURE_YOSHI = 276
    YOSHI_EGG = 277
    CAPTURE_KOOPA = 278
    CAPTURE_DAMAGE_KOOPA = 279
    CAPTURE_WAIT_KOOPA = 280
    THROWN_KOOPA_F = 281
    THROWN_KOOPA_B = 282
    CAPTURE_KOOPA_AIR = 283
    CAPTURE_DAMAGE_KOOPA_AIR = 284
    CAPTURE_WAIT_KOOPA_AIR = 285
    THROWN_KOOPA_AIR_F = 286
    THROWN_KOOPA_AIR_B = 287
    CAPTURE_KIRBY = 288
    CAPTURE_WAIT_KIRBY = 289
    THROWN_KIRBY_STAR = 290
    THROWN_COPY_STAR = 291
    THROWN_KIRBY = 292
    BARREL_WAIT = 293
    BURY = 294
    BURY_WAIT = 295
    BURY_JUMP = 296
    DAMAGE_SONG = 297
    DAMAGE_SONG_WAIT = 298
    DAMAGE_SONG_RV = 299
    DAMAGE_BIND = 300
    CAPTURE_MEWTWO = 301
    CAPTURE_MEWTWO_AIR = 302
    THROWN_MEWTWO = 303
    THROWN_MEWTWO_AIR = 304
    WARP_STAR_JUMP = 305
    WARP_STAR_FALL = 306
    HAMMER_WAIT = 307
    HAMMER_WALK = 308
    HAMMER_TURN = 309
    HAMMER_KNEE_BEND = 310
    HAMMER_FALL = 311
    HAMMER_JUMP = 312
    HAMMER_LANDING = 313
    KINOKO_GIANT_START = 314
    KINOKO_GIANT_START_AIR = 315
    KINOKO_GIANT_END = 316
    KINOKO_GIANT_END_AIR = 317
    KINOKO_SMALL_START = 318
    KINOKO_SMALL_START_AIR = 319
    KINOKO_SMALL_END = 320
    KINOKO_SMALL_END_AIR = 321
    ENTRY = 322
    ENTRY_START = 323
    ENTRY_END = 324
    DAMAGE_ICE = 325
    DAMAGE_ICE_JUMP = 326
    CAPTURE_MASTERHAND = 327
    CAPTUREDAMAGE_MASTERHAND = 328
    CAPTUREWAIT_MASTERHAND = 329
    THROWN_MASTERHAND = 330
    CAPTURE_KIRBY_YOSHI = 331
    KIRBY_YOSHI_EGG = 332
    CAPTURE_LEADEAD = 333
    CAPTURE_LIKELIKE = 334
    DOWN_REFLECT = 335
    CAPTURE_CRAZYHAND = 336
    CAPTUREDAMAGE_CRAZYHAND = 337
    CAPTUREWAIT_CRAZYHAND = 338
    THROWN_CRAZYHAND = 339
    BARREL_CANNON_WAIT = 340
    WAIT1 = 341
    WAIT2 = 342
    WAIT3 = 343
    WAIT4 = 344
    WAIT_ITEM = 345
    SQUAT_WAIT1 = 346
    SQUAT_WAIT2 = 347
    SQUAT_WAIT_ITEM = 348
    GUARD_DAMAGE = 349
    ESCAPE_N = 350
    ATTACK_S4HOLD = 351
    HEAVY_WALK1 = 352
    HEAVY_WALK2 = 353
    ITEM_HAMMER_WAIT = 354
    ITEM_HAMMER_MOVE = 355
    ITEM_BLIND = 356
    DAMAGE_ELEC = 357
    FURA_SLEEP_START = 358
    FURA_SLEEP_LOOP = 359
    FURA_SLEEP_END = 360
    WALL_DAMAGE = 361
    CLIFF_WAIT1 = 362
    CLIFF_WAIT2 = 363
    SLIP_DOWN = 364
    SLIP = 365
    SLIP_TURN = 366
    SLIP_DASH = 367
    SLIP_WAIT = 368
    SLIP_STAND = 369
    SLIP_ATTACK = 370
    SLIP_ESCAPE_F = 371
    SLIP_ESCAPE_B = 372
    APPEAL_S = 373
    ZITABATA = 374
    CAPTURE_KOOPA_HIT = 375
    THROWN_KOOPA_END_F = 376
    THROWN_KOOPA_END_B = 377
    CAPTURE_KOOPA_AIR_HIT = 378
    THROWN_KOOPA_AIR_END_F = 379
    THROWN_KOOPA_AIR_END_B = 380
    THROWN_KIRBY_DRINK_S_SHOT = 381
    THROWN_KIRBY_SPIT_S_SHOT = 382


class CSSCharacter(IntEnum):
    CAPTAIN_FALCON = 0
    DONKEY_KONG = 1
    FOX = 2
    GAME_AND_WATCH = 3
    KIRBY = 4
    BOWSER = 5
    LINK = 6
    LUIGI = 7
    MARIO = 8
    MARTH = 9
    MEWTWO = 10
    NESS = 11
    PEACH = 12
    PIKACHU = 13
    ICE_CLIMBERS = 14
    JIGGLYPUFF = 15
    SAMUS = 16
    YOSHI = 17
    ZELDA = 18
    SHEIK = 19
    FALCO = 20
    YOUNG_LINK = 21
    DR_MARIO = 22
    ROY = 23
    PICHU = 24
    GANONDORF = 25
    MASTER_HAND = 26
    WIREFRAME_MALE = 27
    WIREFRAME_FEMALE = 28
    GIGA_BOWSER = 29
    CRAZY_HAND = 30
    SANDBAG = 31
    POPO = 32

    @classmethod
    def from_internal_id(cls, internal_id):
        char = InGameCharacter(internal_id)
        if char is InGameCharacter.NANA or char is InGameCharacter.POPO:
            return cls.ICE_CLIMBERS
        else:
            return cls[char.name]

    def __repr__(self):
        return '%s.%s' % (self.__class__.__name__, self.name)


class InGameCharacter(IntEnum):
    MARIO = 0
    FOX = 1
    CAPTAIN_FALCON = 2
    DONKEY_KONG = 3
    KIRBY = 4
    BOWSER = 5
    LINK = 6
    SHEIK = 7
    NESS = 8
    PEACH = 9
    POPO = 10
    NANA = 11
    PIKACHU = 12
    SAMUS = 13
    YOSHI = 14
    JIGGLYPUFF = 15
    MEWTWO = 16
    LUIGI = 17
    MARTH = 18
    ZELDA = 19
    YOUNG_LINK = 20
    DR_MARIO = 21
    FALCO = 22
    PICHU = 23
    GAME_AND_WATCH = 24
    GANONDORF = 25
    ROY = 26
    MASTER_HAND = 27
    CRAZY_HAND = 28
    WIREFRAME_MALE = 29
    WIREFRAME_FEMALE = 30
    GIGA_BOWSER = 31
    SANDBAG = 32


class Stage(IntEnum):
    FOUNTAIN_OF_DREAMS = 2
    POKEMON_STADIUM = 3
    PRINCESS_PEACHS_CASTLE = 4
    KONGO_JUNGLE = 5
    BRINSTAR = 6
    CORNERIA = 7
    YOSHIS_STORY = 8
    ONETT = 9
    MUTE_CITY = 10
    RAINBOW_CRUISE = 11
    JUNGLE_JAPES = 12
    GREAT_BAY = 13
    HYRULE_TEMPLE = 14
    BRINSTAR_DEPTHS = 15
    YOSHIS_ISLAND = 16
    GREEN_GREENS = 17
    FOURSIDE = 18
    MUSHROOM_KINGDOM_I = 19
    MUSHROOM_KINGDOM_II = 20
    VENOM = 22
    POKE_FLOATS = 23
    BIG_BLUE = 24
    ICICLE_MOUNTAIN = 25
    ICETOP = 26
    FLAT_ZONE = 27
    DREAM_LAND_N64 = 28
    YOSHIS_ISLAND_N64 = 29
    KONGO_JUNGLE_N64 = 30
    BATTLEFIELD = 31
    FINAL_DESTINATION = 32



import io, sys, ubjson
from datetime import datetime
from enum import Enum

import slippi.event as evt
from slippi.id import InGameCharacter
from slippi.util import *


class Game:
    """Replay data from a game of Super Smash Brothers Melee."""

    def __init__(self, path):
        """Reads data from the Slippi (.slp) replay file at `path`."""

        self.metadata = None
        """:py:class:`Metadata`: Miscellaneous data relevant to the game but not directly provided by Melee"""

        self.start = None
        """:py:class:`slippi.event.Start`: Information about the start of the game"""

        self.frames = []
        """list(:py:class:`slippi.event.Frame`): Every frame of the game, indexed by frame number (omits frames that occur before the timer starts)"""

        self.end = None
        """:py:class:`slippi.event.End`: Information about the end of the game"""

        self._out_of_order = False

        self._parse_file(path)

    def _parse_event_payloads(self, stream):
        (code, payload_size) = unpack('BB', stream)
        event_type = evt.EventType(code)
        if event_type is not evt.EventType.EVENT_PAYLOADS:
            raise Exception('expected event payloads, but got: %s' % event_type)
        payload_size -= 1 # includes size byte for some reason
        command_count = payload_size // 3
        if command_count * 3 != payload_size:
            raise Exception('payload size not divisible by 3: %d' % payload_size)
        payload_sizes = {}
        for i in range(command_count):
            (code, size) = unpack('BH', stream)
            payload_sizes[evt.EventType(code)] = size
        return payload_sizes

    def _parse_event(self, event_stream, payload_sizes):
        (code,) = unpack('B', event_stream)
        try:
            event_type = evt.EventType(code)
        except ValueError:
            print('skipping unknown event code: %d' % code)
            return None
        payload = event_stream.read(payload_sizes[event_type])
        stream = io.BytesIO(payload)
        if event_type is evt.EventType.GAME_START:
            return evt.Start._parse(stream)
        elif event_type is evt.EventType.PRE_FRAME:
            return evt.Frame.Event(evt.Frame.Event.Id(stream),
                                   evt.Frame.Port.Data.Pre(stream))
        elif event_type is evt.EventType.POST_FRAME:
            return evt.Frame.Event(evt.Frame.Event.Id(stream),
                                   evt.Frame.Port.Data.Post(stream))
        elif event_type is evt.EventType.GAME_END:
            return evt.End._parse(stream)
        else:
            raise Exception('unexpected event: %s' % event_type)

    def _parse_file(self, path):
        """Parses the .slp file at `path`. Called automatically by our constructor."""

        with open(path, 'rb') as f:
            json = ubjson.load(f)

        self.metadata = self.Metadata._parse(json['metadata'])

        stream = io.BytesIO(json['raw'])
        payload_sizes = self._parse_event_payloads(stream)

        try:
            while True:
                event = self._parse_event(stream, payload_sizes)
                if isinstance(event, evt.Frame.Event):
                    if event.id.frame < 0:
                        continue

                    if not self._out_of_order and len(self.frames) != event.id.frame and len(self.frames) != event.id.frame + 1:
                        self._out_of_order = True
                        print('out-of-order frame: %d' % event.id.frame, file=sys.stderr)

                    while (len(self.frames) <= event.id.frame):
                        if self.frames:
                            self.frames[-1]._finalize()
                        self.frames.append(evt.Frame())

                    port = self.frames[event.id.frame].ports[event.id.port]
                    if not port:
                        port = evt.Frame.Port()
                        self.frames[event.id.frame].ports[event.id.port] = port

                    if event.id.is_follower:
                        port.follower = evt.Frame.Port.Data()
                        data = port.follower
                    else:
                        data = port.leader

                    if isinstance(event.data, evt.Frame.Port.Data.Pre):
                        data.pre = event.data
                    else:
                        data.post = event.data
                elif isinstance(event, evt.Start):
                    self.start = event
                elif isinstance(event, evt.End):
                    self.end = event
                else:
                    raise Exception('unexpected event: %s' % event)
        except EofException:
            pass

    def __repr__(self):
        return '<%s: {metadata: %s, start: %s, end: %s, frames: [...]}>' % \
            (self.__class__.__qualname__, self.metadata, self.start, self.end)


    class Metadata:
        """Miscellaneous data relevant to the game but not directly provided by Melee."""
        def __init__(self, date, duration, platform, players):
            self.date = date #: :py:class:`datetime`: Game start date & time
            self.duration = duration #: :py:class:`int`: Duration of game, in frames
            self.platform = platform #: :py:class:`Platform`: Platform the game was played on (console/dolphin)
            self.players = players #: tuple(optional(:py:class:`Player`)): Player metadata by port (port 1 is at index 0; empty ports will contain None)

        @classmethod
        def _parse(cls, json):
            d = json['startAt']
            if d[-1] == 'Z':
                d = d[:-1] + '+0000'
            date = datetime.strptime(d, '%Y-%m-%dT%H:%M:%S%z')
            try:
                duration = 1 + json['lastFrame']
            except KeyError: duration = None
            platform = cls.Platform(json['playedOn'])
            players = [None, None, None, None]
            for i in range(4):
                try:
                    players[i] = cls.Player._parse(json['players'][str(i)]['characters'])
                except KeyError: pass
            return cls(date=date, duration=duration, platform=platform, players=tuple(players))

        __repr__ = dict_repr

        def __eq__(self, other):
            if not isinstance(other, self.__class__):
                return NotImplemented
            return self.date == other.date and self.duration == other.duration and self.platform == other.platform and self.players == other.players


        class Player:
            def __init__(self, characters):
                self.characters = characters #: dict(:py:class:`slippi.id.InGameCharacter`, :py:class:`int`): Character(s) used, with usage duration in frames (for Zelda/Sheik)

            @classmethod
            def _parse(cls, json):
                characters = {}
                for char_id, duration in json.items():
                    characters[InGameCharacter(int(char_id))] = duration - 123 # ignore 123 frames before timer starts
                return cls(characters)

            __repr__ = dict_repr

            def __eq__(self, other):
                if not isinstance(other, self.__class__):
                    return NotImplemented
                return self.characters == other.characters


        class Platform(Enum):
            CONSOLE = 'console'
            DOLPHIN = 'dolphin'
